/*Forrest Heller 2008*/
#include "derivative.h"
#include "counter.h"
#include <stdlib.h>
/*the design:
counting_init() will be called to set up a game_modes array with randomized parameters

counting_loop() will be called at a consistent period of time
It does:
It counts a delay.  Once the delay has reached a certain limit the mode function (based on the current game_mode in the
game_modes array) will be called

Then take a status flag and determine action
*/



#define MAX_MODES 5
#define BASE_NUMBER 4
#define MAX_NUMBER OUTPUT_MASK
#define S_INIT 0
#define S_RUNNING 1
#define MAX_RUN_COUNT 5
#define MODE_FUNCTIONS 5 //sizeof(mode_funcs)/sizeof(unsigned char (*)(void))
#define FAST_DELAY_VALUE 50
#define SLOW_DELAY_VALUE 255
#define MAX_RANDOM_RUNS 5
#define MIN_RANDOM_RUNS 3
#define ZERO_AT_RANDOM_CHANCE 5
#define ZERO_AT_END_CHANCE 20
#define F_FAST 1
#define F_ZERO_AT_END 2
#define F_ZERO_AT_RAND 4
#define PLAYER_1 0X20
#define PLAYER_2 0X01

/*state flags*/
unsigned char number;
unsigned char zero_flag;
unsigned char rollover;
/*the current mode*/
unsigned char mode;
/*the current state of the game -- counting_loop();*/
unsigned char state;

/*this specifies the number of times to run randomly
It is used by the mode_random mode function*/
unsigned char random_mode_run_times;
/*this is the status flag that is used
See the mode functions for more detail*/
unsigned char mode_status_flag = 0;

/*function pointer table to mode functions*/
void (*mode_funcs[])() = {
	mode_count_down,
	mode_count_up,
	mode_random,
	mode_two_step_down,
	mode_two_step_up
};

/*this is the maximum run count for everything--it is determined that when this reaches 0 there are no more 
modes to run see the loop in count_loop()*/
unsigned char total_run_count;

/*this is set up by the counting_init function with randomized parameters*/
struct game_mode game_modes[MAX_MODES];

unsigned char game_modes_count; //number of game modes
unsigned char current_game_mode; //current pointer to game mode

/*this delay is used in counting_loop to determine when to run a mode functions in game_modes*/
unsigned char delay = 0;


extern unsigned char random_seed;
extern unsigned char game_mode;
unsigned char prand(unsigned char max) { 
	asm {
		;I stole this algorithm from http://www.piclist.com/techref/microchip/rand8bit.htm and translated it
		lda   random_seed
		RORA  
		BCC   no_xor
		EOR   #180
	no_xor:
		sta   random_seed
	}
	return random_seed % max;
}

/*
chance = 0 to 255 (0 to 100%) that will you get value returned*/
unsigned char weighted_rand(unsigned char chance, unsigned char value) {
	
	if (prand(255) < chance) {
		return value;
	} else {
		return 0;
	}
}
/*this will set up the data structure needed to run the game*/
void counting_init() {
	unsigned char i;
	game_modes_count = prand(MAX_MODES) + 1;
	total_run_count = 0;
	for(i = 0; i < game_modes_count; i++) {
		game_modes[i].mode = mode_funcs[prand(MODE_FUNCTIONS)];
		/*will we be in FaSt moDe*/
		game_modes[i].flags = weighted_rand(127,F_FAST);
		game_modes[i].run_count = prand(MAX_RUN_COUNT) + 1;
		/*stop at random?*/
		game_modes[i].flags |= weighted_rand(5*(i*2+1),F_ZERO_AT_RAND);
		/*end at random?*/
		game_modes[i].flags |= weighted_rand(5*(i*2+1),F_ZERO_AT_END);
		/*append to the total count*/
		total_run_count += game_modes[i].run_count;
	}  
	/*at the end of the game we want a zero flag*/
	i--;
	game_modes[i].flags |= F_ZERO_AT_END;
	
	
	number = MAX_NUMBER & (prand(MAX_NUMBER - BASE_NUMBER) + BASE_NUMBER);
	state = S_RUNNING;  //weiner
	current_game_mode = 0;
	zero_flag = 0;
	delay = 0;
	rollover = 255;
	/*mode-specific flags*/
	random_mode_run_times = prand(MAX_RUN_COUNT);
	mode_status_flag = 1;
	
}
unsigned char winner = 0;
unsigned char counting_loop() {
	
		delay++;
		/*is the fast flag turned on?
		If it is, then see if delay meets the fast criteria
		If not, then see if the delay meets the slow criteria
		If either are met, run the game_modes current mode function and reset delay*/
		if ((game_modes[current_game_mode].flags & F_FAST) == F_FAST) {
			if (delay >= FAST_DELAY_VALUE) {
				if ((game_modes[current_game_mode].flags & F_ZERO_AT_RAND) == F_ZERO_AT_RAND) {
					if (prand(255) < ZERO_AT_RANDOM_CHANCE) {
						number  = 0;
					} 
				}
				if ((game_modes[current_game_mode].flags & F_ZERO_AT_END) == F_ZERO_AT_END) {
					if (prand(255) < ZERO_AT_END_CHANCE) {
						zero_flag = 1;
					} 
				}
				game_modes[current_game_mode].mode();
				delay = 0;
			}
		} else {
			if (delay >= SLOW_DELAY_VALUE) {					
  				if ((game_modes[current_game_mode].flags & F_ZERO_AT_RAND) == F_ZERO_AT_RAND) {
					if (prand(255) < ZERO_AT_RANDOM_CHANCE) {
						number  = 0;
					} 
				}
				if ((game_modes[current_game_mode].flags & F_ZERO_AT_END) == F_ZERO_AT_END) {
					if (prand(255) < ZERO_AT_END_CHANCE) {
						zero_flag = 1;
					} 
				}
				game_modes[current_game_mode].mode();
				delay = 0;
			} 
		}
		/*now we have a status indicator in the game mode functions
		'lets do something with it
		and then reset it to 0*/
		switch(mode_status_flag) {
		case 0:
			number = 0;
			mode_status_flag  = 1;
		break;
		case 2:
			
			/*select random game mode*/
			
			/*if every game mode has run then...make a new one*/
			total_run_count--;
			if (total_run_count == 0) {
				/*LAST!*/
				number = 0;
				break;
			}
			game_modes[current_game_mode].run_count--;
			/*while a game mode's run count is not 0, then continue selecting*/
			
			do {
				current_game_mode = prand(game_modes_count);
			} while (game_modes[current_game_mode].run_count == 0);				
		default:
			/*reset the mode status flag*/
			mode_status_flag = 1;
			break;
			
		}
		asm {
			lda PTAD
			and  #INPUT_MASK
			;did someone press something?
			TSTA 
			BEQ load_number
		}
		/*Whoever pressed the button first lost!*/
		if ((PTAD & PLAYER_1) == PLAYER_1) {
			winner = PLAYER_2; 
		}  else {
			winner = PLAYER_1; 
		}
		
		asm {
			jmp   end_game
		load_number:
			eor  number
			;if (number != 0) goto skip_zero
				TSTA
			bne    skip_zero				
			
			;by doing this here I elimnate having to check number after
			;storing PTAD 0
			sta   PTAD ;I can take advantage of the fact that here  the zero bit is set
			;I don't use a compare and branch if equal because I am interested in the value of PTAD
		zero_loop:
			
			lda PTAD 
			AND #INPUT_MASK
			TSTA
			BEQ zero_loop
		game_winner:
			;store winner
			sta winner
		end_game:
		}
		zero_flag = 0;
		/*re-use variables delay and mode_status_flag
		are not important anymore*/
		for(delay = 0; delay < 100; delay++) {
		  mode_count_up();
		  PTAD = number;
		  delay_good();
		}

		if ((winner & PLAYER_2) == PLAYER_2) {
			PTAD = PNUMBER(2);
		} else if ((winner & PLAYER_1) == PLAYER_1) {
			PTAD = PNUMBER(1);
		}
		delay_25s();
		delay_25s();
		delay_25s();
		delay_25s();
		/*wait until all buttons turned off*/
		while((PTAD & NOUTPUT_MASK) != 0) random_seed++;
		/*set game mode in counter_main.c*/
		game_mode = 0;
		return 0;
		asm {
			
		skip_zero:
			sta  PTAD ;I can take advantage of the fact that here  the zero bit is set
			
			
		}
}
/*these are the modes
each mode is called from the counting_loop on the appropriate delay
They are assigned during counting_init() to a game mode in the game_modes array
They will:
modify number
set a mode_status_flag and return
*/
void mode_count_down() {
	/*check for rollover condition -- notice that I am checking it to see if it IS GOING TO GO TO ZERO
	and NOT IF IT IS AT 0*/
	if (number == BASE_NUMBER || number == rollover) {
		/*check to see if zero_flag is not set*/
		if (zero_flag == 0) {
			/*rollover to 1*/
			number = MAX_NUMBER;
			mode_status_flag = 2;
		} else {
			/*we return 0*/	
			number = 0;
			mode_status_flag = 0;
		}
		return;
	}
	mode_status_flag = 1;
	/*apply action*/
	number -= BASE_NUMBER;
}
void mode_count_up() {
	/*check for rollover condition*/
	if (number == MAX_NUMBER || number == rollover) {
		/*check to see if zero_flag is not set*/
		if (zero_flag == 0) {
			/*rollover to 1*/
			number = BASE_NUMBER;
			mode_status_flag = 2;
		} else {               
			/*we return 0*/
			number = 0;
			mode_status_flag = 0;
		}
		return;
	}
	/*apply action*/
	number += BASE_NUMBER;
	
	mode_status_flag = 1;
}
void mode_two_step_up() {
	/*check for rollover condition*/
	if (number >= (MAX_NUMBER-BASE_NUMBER) || number == rollover) {
		/*check to see if zero_flag is not set*/
		if (zero_flag == 0) {
			/*rollover to 1*/
			number = BASE_NUMBER;
			mode_status_flag = 2;
		} else {               
			/*we return 0*/
			number = 0;
			mode_status_flag = 0;
		}
		return;
	}
	/*apply action*/
	number += (2*BASE_NUMBER);
	mode_status_flag = 1;
}
void mode_two_step_down() {
	/*check for rollover condition*/
	if (number <= (BASE_NUMBER*2) || number == rollover) {
		/*check to see if zero_flag is not set*/
		if (zero_flag == 0) {
			/*rollover to 1*/
			number = MAX_NUMBER;
			mode_status_flag = 2;
			return;
		} else {               
			/*we return 0*/
			number = 0;
			mode_status_flag = 0;
			return;
		}
	}
	/*apply action*/
	number -= (2*BASE_NUMBER);
	mode_status_flag = 1;
}
void mode_random() {
	/*number of random times to run*/
	static unsigned char run_times = 0;
	
	/*check for rollover condition*/
	if (number == rollover || number == 0 || ++run_times == random_mode_run_times) {
		/*check to see if zero_flag is not set*/
		if (zero_flag == 0) {
			/*more randomer*/
			random_mode_run_times = prand(MAX_RANDOM_RUNS)+ MIN_RANDOM_RUNS;
			run_times = 0;
			mode_status_flag = 2;
			if (number == 0) number = PNUMBER(5);
		} else {               
			/*we return 0*/
			number = 0;
			mode_status_flag = 0;
		}
		return;
		
	}
	/*apply action*/
	number = prand(255) & MAX_NUMBER; 
	if (number== 0) {
		number = BASE_NUMBER;
	}
	mode_status_flag = 1;
}
void delay_good() {
	/*If one cared, one could combine these to lesser writes*/
	MTIMCLK_PS = 8; /*128 prescale*/
	MTIMCLK_CLKS = 0; /*internal clock source*/
	MTIMSC_TOF = 0;
	MTIMSC_TOIE = 0;
	MTIMSC_TSTP = 0;    
	while (MTIMSC_TOF == 0);
	
}
void delay_25s() {
	unsigned char i;
	for( i = 0; i < 15; i++) {
		/*If one cared, one could combine these to lesser writes*/
		MTIMCLK_PS = 8; /*256 prescale*/
		MTIMCLK_CLKS = 0; /*internal clock source*/
		MTIMSC_TOF = 0;
		MTIMSC_TOIE = 0;
		MTIMSC_TSTP = 0;    
		while (MTIMSC_TOF == 0);
	}
	
}